# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gaphas']

package_data = \
{'': ['*']}

install_requires = \
['PyGObject>=3.20.0,<4.0.0',
 'importlib_metadata>=1.3.0,<2.0.0',
 'pycairo>=1.13.0,<2.0.0']

setup_kwargs = {
    'name': 'gaphas',
    'version': '2.1.2',
    'description': 'Gaphas is a GTK+ based diagramming widget',
    'long_description': '# Gaphas\n[![Build state](https://github.com/gaphor/gaphas/workflows/build/badge.svg)](https://github.com/gaphor/gaphas/actions)\n[![Maintainability](https://api.codeclimate.com/v1/badges/e9837cc647b72119fd11/maintainability)](https://codeclimate.com/github/gaphor/gaphas/maintainability)\n[![Test Coverage](https://api.codeclimate.com/v1/badges/e9837cc647b72119fd11/test_coverage)](https://codeclimate.com/github/gaphor/gaphas/test_coverage)\n![Docs build state](https://readthedocs.org/projects/gaphas/badge/?version=latest)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n[![standard-readme compliant](https://img.shields.io/badge/readme%20style-standard-brightgreen.svg?style=flat)](https://github.com/RichardLitt/standard-readme)\n[![Gitter](https://img.shields.io/gitter/room/nwjs/nw.js.svg)](https://gitter.im/Gaphor/Lobby?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge&utm_content=badge)\n[![All Contributors](https://img.shields.io/badge/all_contributors-9-orange.svg?style=flat-square)](#contributors)\n\n> Gaphas is the diagramming widget library for Python.\n\n![Gaphas Demo](docs/images/gaphas-demo.gif)\n\nGaphas is a library that provides the user interface component (widget) for drawing diagrams. Diagrams can be drawn to screen and then easily exported to a variety of formats, including SVG and PDF. Want to build an app with chart-like diagrams? Then Gaphas is for you! Use this library to build a tree, network, flowchart, or other diagrams.\n\nThis library is currently being used by [Gaphor](https://github.com/gaphor/gaphor) for UML drawing,\n[RAFCON](https://github.com/DLR-RM/RAFCON) for state-machine based robot control, and [ASCEND](http://ascend4.org/) for solving mathematical models.\n\n## :bookmark_tabs: Table of Contents\n\n- [Background](#background)\n- [Install](#install)\n- [Usage](#usage)\n- [API](#api)\n- [Contributing](#contributing)\n- [License](#license)\n\n## :scroll: Background\n\nGaphas was built to provide the foundational diagramming portions of [Gaphor](https://github.com/gaphor/gaphor). Since Gaphor is built on GTK and Cairo, [PyGObject](https://pygobject.readthedocs.io/) provides access to the GUI toolkit and [PyCairo](https://pycairo.readthedocs.io/) to the 2D graphics library. However, there wasn\'t a project that abstracted these technologies to easily create a diagramming tool. Hence, Gaphas was created as a library to allow others to create a diagramming tool using GTK and Cairo.\n\nHere is how it works:\n\n- Items (Canvas items) can be added to a Canvas.\n- The Canvas maintains the tree structure (parent-child relationships between items).\n- A constraint solver is used to maintain item constraints and inter-item constraints.\n- The item (and user) should not be bothered with things like bounding-box calculations.\n- Very modular--e.g., handle support could be swapped in and swapped out.\n- Rendering using Cairo.\n\nThe main portions of the library include:\n\n- canvas - The main canvas class (container for Items).\n- items - Objects placed on a Canvas.\n- solver - A constraint solver to define the layout and connection of items.\n- view - Responsible for the calculation of bounding boxes which is stored in a quadtree data structure for fast access.\n- gtkview - A view to be used in GTK applications that interacts with users with tools.\n- painters - The workers used to paint items.\n- tools - Tools are used to handle user events (such as mouse movement and button presses).\n- aspects - Provides an intermediate step between tools and items.\n\n## :floppy_disk: Install\n\nTo install Gaphas, simply use pip:\n\n```bash\n$ pip install gaphas\n```\n\nUse of a\n[virtual environment](https://virtualenv.pypa.io/en/latest/) is highly recommended.\n\n### Development\n\nTo setup a development environment with Linux:\n\n```bash\n$ sudo apt-get install -y python3-dev python3-gi python3-gi-cairo\n    gir1.2-gtk-3.0 libgirepository1.0-dev libcairo2-dev\n$ pip install poetry\n$ poetry install\n```\n\n## :flashlight: Usage\n\n```python\nimport gi\ngi.require_version("Gtk", "3.0")\nfrom gi.repository import Gtk\n\nfrom gaphas import Canvas, GtkView\nfrom gaphas.examples import Box\nfrom gaphas.painter import DefaultPainter\nfrom gaphas.item import Line\nfrom gaphas.segment import Segment\n\n\ndef create_canvas(canvas, title):\n    # Setup drawing window\n    view = GtkView()\n    view.painter = DefaultPainter()\n    view.canvas = canvas\n    window = Gtk.Window()\n    window.set_title(title)\n    window.set_default_size(400, 400)\n    win_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)\n    window.add(win_box)\n    win_box.pack_start(view, True, True, 0)\n\n    # Draw gaphas box\n    b2 = Box(60, 60)\n    b2.min_width = 40\n    b2.min_height = 50\n    b2.matrix.translate(170, 170)\n    canvas.add(b2)\n\n    # Draw gaphas line\n    line = Line()\n    line.matrix.translate(100, 60)\n    canvas.add(line)\n    line.handles()[1].pos = (30, 30)\n    segment = Segment(line, view=None)\n    segment.split_segment(0)\n\n    window.show_all()\n    window.connect("destroy", Gtk.main_quit)\n\n\nc = Canvas()\ncreate_canvas(c, "Simple Gaphas App")\nGtk.main()\n```\n\n### Overview\n\nThe `Canvas` class acts as a container for Item\'s . The item\'s parent/child relationships are maintained here, not in the `Item`.\n\nAn `Item` can have a set of handles which can be used to manipulate the item (although this is not necessary). Each item has its own coordinate system with x and y position, for example, a (0, 0) point `Item.matrix` is the transformation relative to the parent item of the `Item`, as defined in the `Canvas`.\n\nHandles can connect to [ports](https://gaphas.readthedocs.io/en/latest/ports.html). A `Port` is a location (line or point) where a handle is allowed to connect on another\nitem. The process of connecting depends on the case at hand, but most often involves the creation of some sort of constraint between the Handle and the item it is connecting to.\n\nThe Canvas also contains a [constraint solver](https://gaphas.readthedocs.io/en/latest/solver.html) (from solver.py) that can be used to solve mathematical dependencies\nbetween items (such as Handles that should be aligned). The constraint solver can also be used to keep constraints contained within the item satisfied, for example, to make sure a box maintains its rectangular shape.\n\nA View is used to visualize a Canvas. On a View, a Tool (from tool.py) can be applied, which will handle user input like button and key presses. Painters (from painter.py) are used to do the actual drawing. This module also makes it easy to draw to other media other than a screen, such as a printer or a PDF document.\n\n### Updating Item state\n\nIf items need updating, it sends out an update request on the Canvas (`Canvas.request_update()`). The canvas\nperforms an update by performing the following steps:\n\n1. Pre-update using `Item.pre_update(context)` for each item marked for update.\n2. Update the Canvas-to-Item matrices, for fast transformation of coordinates from the Canvas\' to the items\' coordinate      system.\n3. Solve the constraints.\n4. Normalize the items by setting the coordinates of the first handle to (0, 0).\n5. Update the Canvas-to-Item matrices for items that have been changed by normalization.\n6. Post-update using `Item.post_update(context)` for each item marked for update, including items that have been\n   marked during the constraint solving step.\n\nGaphas attempts to do as much updating as possible in the {pre|post}_update() methods, since they are called when the application is not handling user input.\n\nThe context contains a CairoContext. This can be used, for example, to calculate the dimensions of text. One thing to keep in mind is that updating is done from the Canvas. Items should not update sub-items. After the update steps are complete, the Item should be ready to be drawn.\n\n### Constraint solving\n\nConstraint solving is one of the big features of this library. The Solver is able to mathematically solve these constraint rules that are applied to an item or between items. Constraints are applied to items through variables owned by\nthe item. An example of applying a constraint to an item is that Element items use constraints to maintain their rectangular shape. An example of applying constraints between items is to apply a constraint between a line and a box in\norder to connect them.\n\nConstraints that apply to one item are pretty straight forward, as all variables live in the same coordinate system of the item. The variables, like the Handle\'s x and y coordinate can simply be put in a constraint.\n\nWhen two items are connected to each other and constraints are created, a problem shows up: variables live in separate coordinate systems. In order to overcome this problem, a Projection (from solver.py) has been defined. With a\nProjection instance, a variable can be "projected" on another coordinate system. In this case, the Canvas\' coordinate system is used when two items are connected to each other.\n\n### Drawing\n\nDrawing is done by the View. All items marked for redraw, the items that have been updated, will be drawn in the order in which they reside in the Canvas. The order starts with the first root item, then its children, then second root item, etc.\n\nThe view context passed to the Items draw() method has the following properties:\n\n- view - The view we\'re drawing to.\n- cairo - The CairoContext to draw to.\n- selected - True if the item is actually selected in the view.\n- focused - True if the item has the focus\n- hovered - True if the mouse pointer if over the item. Only the top-most item is marked as hovered.\n- dropzone - The item is marked as the drop zone. When this happens then an item is dragged over the item, and if it is     dropped, it will become a child of this item.\n- draw_all - True if everything drawable on the item should be drawn, for example, when calculating the bounding boxes of   an item.\n\nThe View automatically calculates the bounding box for the item, based on the items drawn in the draw (context) function (this is only done when really necessary, e.g., after an update of the item). The bounding box is in viewport coordinates.\n\nThe actual drawing is done by Painters (painter.py). A series of Painters have been defined: one for handles, one for items, etc.\n\n### Tools\n\nBehaviour is added to the Canvas (view) by tools. Tools can be chained together in order to provide more complex behaviour.\n\nTo make it easy, a DefaultTool has been defined which is a ToolChain instance with the tools added as follows:\n\n- ToolChain - Delegates to a set of individual tools and keeps track of which tool has grabbed the focus. This normally     happens when the user presses a mouse button. Once this happens, the tool requests a "grab" and all events, like motion   or button release, are sent directly to the focused tool.\n\n- HoverTool - Makes the item under the mouse button the "hovered item." When such an item is drawn, its context.hovered_item flag will be set to True.\n\n- HandleTool - Allows for handles to be dragged around and focuses the item when its handle is clicked on.\n\n- ItemTool - Selects items and enables dragging items around.\n\n- TextEditTool - A demo tool that features a text edit pop-up.\n\n- RubberbandTool - Invoked when the mouse button is pressed on a section of the view where no items or handles are present. It allows the user to select items using a "rubber band" selection box.\n\n### Interaction\n\nTools handle Interaction with the Canvas view (visual component). Although the default tools do a fair amount of work, in most cases you\'ll desire to create some custom connection behaviour. In order to implement these, HandleTool provides hooks including connect, disconnect, and glue.\n\nOne of the challenges you\'ll likely face is what to do when an item is removed from the Canvas and there are other items (lines) connected to it. Gaphas provides a solution to this by providing a disconnect handler to the handle instance once it is connected. A function can be assigned to this disconnect handler, which is then called when the item that is connected to is removed from the Canvas.\n\n### Undo\n\nGaphas has a simple built-in system for registering changes in its classes and notifying the application. This code resides in state.py.\n\nThere is also a "reverter" framework in place. This system is notified when objects change their state, and the framework will figure out the reverse operation that has to be applied in order to undo the operation.\n\n## :mag: API\n\nThe API can be separated into a [Model-View-Controller](https://en.wikipedia.org/wiki/Model%E2%80%93view%E2%80%93controller) with these three parts:\n\n1. The Model, including the canvas and items\n2. The View, called view\n3. The Controller, called tools\n\n### Canvas and Items\n\n#### Class: `gaphas.canvas.Canvas`\n\nThe `Canvas` is a container for items.\n\n```python\ncanvas = Canvas()\n```\n\n#### Class: `gaphas.item.Item`\n\nBase class (or interface) for items on a `Canvas`.\n\n```python\nitem = Item()\n```\n\n##### Properties:\n\n- `matrix`: The item\'s transformation matrix\n- `canvas`: The canvas, which owns an item\n- `constraints`: list of item constraints, automatically registered when the item is added to a canvas; may be extended in subclasses\n\n#### Class: `gaphas.connector.Handle`\n\nHandles are used to support modifications of Items.\n\nIf the handle is connected to an item, the `connected_to` property should refer to the item. A `disconnect` handler should\nbe provided that handles the required disconnect behaviour, for example, cleaning up the constraints and `connected_to`.\n\n- pos (`gaphas.connector.Position`): The position of the item, default value is (0, 0).\n- strength (int): The strength of the handle to use in the constraint solver; default value is NORMAL, which is 20.\n- connectable (bool): Makes the handle connectable to other items; default value is False.\n- movable (bool): Makes the handle moveable; default value is True.\n\n  ```python\n  handle = Handle((10, 10), connectable=True)\n  ```\n\n#### Class: `gaphas.connector.LinePort`\n\nThe Line Port is part of an item that provides a line between two handles.\n\n- start (`gaphas.connector.Position`): The start position of the line.\n- end (`gaphas.connector.Position`): The end position of the line.\n\n```python\np1, p2 = (0.0, 0.0), (100.0, 100.0)\nport = LinePort(p1, p2)\n```\n\n#### Class: `gaphas.connector.PointPort`\n\nThe Point Port connects the handle to an item using a port at the location of the handle.\n\n```python\nh = Handle((10, 10))\nport = PointPort(h.pos)\n```\n\n#### Class: `gaphas.solver.Solver`\n\nA Solver solves constraints.\n\n```python\na, b, c = Variable(1.0), Variable(2.0), Variable(3.0)\nsolver = Solver()\nc_eq = EquationConstraint(lambda a,b: a+b, a=a, b=b)\nsolver.add_constraint(c_eq)\n```\n\n#### Class: `gaphas.constraint.EqualsConstraint`\n\nMake \'a\' and \'b\' equal.\n\n```python\na, b = Variable(1.0), Variable(2.0)\neq = EqualsConstraint(a, b)\neq.solve_for(a)\n```\n\n#### Class: `gaphas.constraint.LessThanConstraint`\n\nEnsure one variable stays smaller than another.\n\n```python\na, b = Variable(3.0), Variable(2.0)\nlt = LessThanConstraint(smaller=a, bigger=b)\nlt.solve_for(a)\n```\n\n#### Class: `gaphas.constraint.CenterConstraint`\n\nEnsures a Variable is kept between two other variables.\n\n```python\na, b, center = Variable(1.0), Variable(3.0), Variable()\neq = CenterConstraint(a, b, center)\neq.solve_for(a)\n```\n\n#### Class: `gaphas.constraint.EquationConstraint`\n\nSolve a linear equation.\n\n```python\na, b, c = Variable(), Variable(4), Variable(5)\ncons = EquationConstraint(lambda a, b, c: a + b - c, a=a, b=b, c=c)\ncons.solve_for(a)\n```\n\n#### Class: `gaphas.constraint.BalanceConstraint`\n\nKeeps three variables in line, maintaining a specific ratio.\n\n```python\na, b, c = Variable(2.0), Variable(3.0), Variable(2.3, WEAK)\nbc = BalanceConstraint(band=(a,b), v=c)\nc.value = 2.4\n```\n\n#### Class: `gaphas.constraint.LineConstraint`\n\nSolves the equation where a line is connected to a line or side at a specific point.\n\n```python\nline = (Variable(0), Variable(0)), (Variable(30), Variable(20))\npoint = (Variable(15), Variable(4))\nlc = LineConstraint(line=line, point=point)\n```\n\n### View\n\n#### Class: `gaphas.view.View`\n\nView class for `gaphas.canvas.Canvas` objects.\n\n```python\ncanvas = Canvas()\nview = View(canvas=canvas)\n```\n\n#### Class: `gaphas.view.GtkView`\n\nGTK+ widget for rendering a `gaphas.canvas.Canvas` to a screen.\n\n```python\ncanvas = Canvas()\nwin = Gtk.Window()\nview = GtkView(canvas=canvas)\nwin.add(view)\n```\n\n#### Class: `gaphas.painter.PainterChain`\n\nChain up a set of painters.\n\n```python\nview.painter = (\n    PainterChain()\n    .append(FreeHandPainter(ItemPainter()))\n    .append(HandlePainter())\n    .append(FocusedItemPainter())\n    .append(ToolPainter())\n)\n```\n\n#### Class: `gaphas.painter.DrawContext`\n\nSpecial context for drawing the item. It contains a cairo context and properties like selected and focused.\n\n- **kwargs: Optional cairo properties for a context.\n\n```python\nDrawContext(\n    painter=self,\n    cairo=cairo,\n    selected=(item in view.selected_items),\n    focused=(item is view.focused_item),\n    hovered=(item is view.hovered_item),\n    dropzone=(item is view.dropzone_item),\n    draw_all=self.draw_all,\n)\n```\n\n#### Class: `gaphas.painter.ItemPainter`\n\nPainter to draw an item.\n\n```python\nsvgview = View(view.canvas)\nsvgview.painter = ItemPainter()\n```\n\n#### Class: `gaphas.painter.CairoBoundingBoxContext`\n\nIt is used to intercept `stroke()`, `fill()`, and other context operations so that the bounding box of the item involved can be calculated.\n\n- cairo (cairo.Context): The cairo context to intercept.\n\n```python\ncairo = CairoBoundingBoxContext(cairo)\n```\n\n#### Class: `gaphas.painter.BoundingBoxPainter`\n\nA type of ItemPainter which is used to calculate the bounding boxes (in canvas coordinates) for the items.\n\n```python\nview.bounding_box_painter = BoundingBoxPainter(ItemPainter())\n```\n\n#### Class: `gaphas.painter.HandlePainter`\n\nDraw handles of items that are marked as selected in the view.\n\n#### Class: `gaphas.painter.ToolPainter`\n\nAllows the Tool defined on a view to conduct drawing.\n\n#### Class: `gaphas.painter.FocusedItemPainter`\n\nUsed to draw on top of all the other layers for the focused item.\n\n### Tools\n\nInteracting with the Canvas is done through tools. Tools tell _what_ has to be done (like moving). To make an element move aspects are defined. Aspects tell how the behaviour has to be performed.\n\n#### Class: `gaphas.tools.ToolChain`\n\nUsed to chain tools together. For example, chain a HoverTool, HandleTool, and SelectionTool in order to combine their functionality in to a new tool.\n\n- view (`gaphas.view.View`): The view to use for the tool chain.\n\n```python\n(ToolChain(view)\n.append(HoverTool())\n.append(ConnectHandleTool())\n.append(PanTool())\n.append(ZoomTool())\n.append(ItemTool())\n.append(TextEditTool())\n.append(RubberbandTool())\n)\n```\n\n#### Class: `gaphas.tools.HoverTool`\n\nMakes the item under the mouse cursor the hovered item.\n\n- view (`gaphas.view.View`): The view to use for the tool; default is None.\n\n#### Class: `gaphas.tools.ItemTool`\n\nDoes selection and dragging of items.\n\n- view (`gaphas.view.View`): The view to use for the tool; default is None.\n\n#### Class: `gaphas.tools.HandleTool`\n\nTool to move handles around.\n\n- view (`gaphas.view.View`): The view to use for the tool; default is None.\n\n#### Class: `gaphas.tools.RubberbandTool`\n\nAllows the user to drag a "rubber band" for selecting items in an area.\n\n- view (`gaphas.view.View`): The view to use for the tool; default is None.\n\n#### Class: `gaphas.tools.PanTool`\n\nCaptures drag events with the middle mouse button and uses them to translate the Canvas within the view.\n\n- view (`gaphas.view.View`): The view to use for the tool; default is None.\n\n#### Class: `gaphas.tools.ZoomTool`\n\nTool for zooming using two different user inputs:\n\n1. Ctrl + middle-mouse dragging in the up and down direction\n2. Ctrl + mouse-wheel\n\n- view (`gaphas.view.View`): The view to use for the tool; default is None.\n\n#### Class: `gaphas.tools.PlacementTool`\n\nTool for placing items on the Canvas.\n\n- view (`gaphas.view.View`): The view to use for the tool.\n- factory (factory object): A Canvas item factory for creating new items.\n- handle_tool (`gaphas.tools.HandleTool`): The handle tool to use.\n- handle_index (int): The index of the handle to be used by the handle tool.\n\n```python\ndef on_clicked(button):\n    view.tool.grab(PlacementTool(view, factory(view, MyLine), HandleTool(), 1))\n```\n\n#### Class: `gaphas.aspects.ItemFinder`\n\nFind an item on the Canvas.\n\n- view (`gaphas.view.View`): The view to use in order to search for an item.\n\n#### Class: `gaphas.aspects.ItemSelection`\n\nManages selection and unselection of items.\n\n- item (`gaphas.item.Item`): The item to set as focused or unfocused.\n- view (`gaphas.view.View`): The view to focus or unfocus on.\n\n#### Class: `gaphas.aspects.ItemInMotion`\n\nManages motion of an item.\n\n- item (`gaphas.item.Item`): The item to move.\n- view (`gaphas.view.View`): The view to to use for move coordinates.\n\n#### Class: `gaphas.aspects.ItemHandleFinder`\n\nFinds handles.\n\n- item (`gaphas.item.Item`): The item.\n- view (`gaphas.view.View`): The view to get the handle at the position from.\n\n```python\ncanvas = Canvas()\nline = Line()\ncanvas.add(line)\nview = View(canvas)\nfinder = HandleFinder(line, view)\n```\n\n#### Class: `gaphas.aspects.ElementHandleSelection`\n\nSelects the handle of a `gaphas.item.Element`.\n\n- item (`gaphas.item.Item`): The Element item that the handle belongs to.\n- handle (`gaphas.connector.Handle`): The handle to select or unselect.\n- view (`gaphas.view.View`): The view that can be used to apply the cursor to.\n\n#### Class: `gaphas.aspects.ItemHandleInMotion`\n\nMove a handle.\n\n- item (`gaphas.item.Item`): The item that the handle belongs to.\n- handle (`gaphas.connector.Handle`): The handle to move.\n- view (`gaphas.view.View`): The view to use for the coordinate system.\n\n#### Class: `gaphas.aspects.ItemConnector`\n\nConnect or disconnect an item\'s handle to another item or port.\n\n- item (`gaphas.item.Item`): The item that the handle belongs to.\n- handle (`gaphas.connector.Handle`): The handle to connect.\n\n#### Class: `gaphas.aspects.ItemConnectionSink`\n\nMakes an item a sink, which is another item that an item\'s handle is connected to like a connected item or port.\n\n- item (`gaphas.item.Item`): The item to look for ports on.\n- port (`gaphas.connector.Port`): The port to use as the sink.\n\n#### Class: `gaphas.aspects.ItemPaintFocused`\n\nPaints on top of all items, just for the focused item and only when it\'s hovered (see\n `gaphas.painter.FocusedItemPainter`).\n\n- item (`gaphas.item.Item`): The focused item.\n- view (`gaphas.view.View`): The view to paint with.\n\n### Extended Behaviour\n\nBy importing the following modules, extra behaviour is added to the default view behaviour.\n\n#### Class: `gaphas.segment.LineSegment`\n\nSplit and merge line segments.\n\n- item (`gaphas.item.Item`): The item of the segment.\n- view (`gaphas.view.View`): The view to use for the split coordinate system.\n\n#### Class: `gaphas.segment.SegmentHandleFinder`\n\nExtends the `gaphas.aspects.ItemHandleFinder` to find a handle on a line, and create a new handle if the mouse is located between two handles. The position aligns with the points drawn by the SegmentPainter.\n\n- item (`gaphas.item.Item`): The item.\n- view (`gaphas.view.View`): The view to get the handle at the position from.\n\n#### Class: `gaphas.segment.SegmentHandleSelection`\n\nExtends the `gaphas.aspects.ItemHandleSelection` to merge segments if the handle is released.\n\n- item (`gaphas.item.Item`): The item that the handle belongs to.\n- handle (`gaphas.connector.Handle`): The handle to select or unselect.\n- view (`gaphas.view.View`): The view to use to apply the cursor to.\n\n#### Class: `gaphas.segment.LineSegmentPainter`\n\nThis painter draws pseudo-handles on a `gaphas.item.Line` by extending `gaphas.aspects.ItemPaintFocused`. Each line can be split by dragging those points, which will result in a new handle.\n\nConnectHandleTool takes care of performing the user interaction required for this feature.\n\n- item (`gaphas.item.Item`): The focused item.\n- view (`gaphas.view.View`): The view to paint with.\n\n#### Class: `gaphas.guide.ElementGuide`\n\nProvides a guide to align items for `gaphas.item.Element`.\n\n- item (`gaphas.item.Item`): The item to provide guides for.\n\n#### Class: `gaphas.guide.LineGuide`\n\nProvides a guide to align items for `gaphas.item.Line`.\n\n- item (`gaphas.item.Item`): The item to provide guides for.\n\n#### Class: `gaphas.guide.GuidedItemInMotion`\n\nMove the item and lock the position of any element that\'s located at the same position.\n\n- item (`gaphas.item.Item`): The item to move.\n- view (`gaphas.view.View`): The view with guides to use for move coordinates.\n\n```python\ncanvas = Canvas()\nview = GtkView(canvas)\nelement = Element()\nguider = GuidedItemInMotion(element, view)\nguider.start_move((0, 0))\n```\n\n#### Class: `gaphas.guide.GuidedItemHandleInMotion`\n\nMove a handle and lock the position of any element that\'s located at the same position.\n\n- item (`gaphas.item.Item`): The item that the handle belongs to.\n- handle (`gaphas.connector.Handle`): The handle to move.\n- view (`gaphas.view.View`): The view with guides to use for the coordinate system.\n\n#### Class: `gaphas.guide.GuidePainter`\n\nPaints on top of all items with guides, just for the focused item and only when it\'s hovered.\n\n- item (`gaphas.item.Item`): The focused item.\n- view (`gaphas.view.View`): The view with guides to paint with.\n\n### Miscellaneous\n\n#### Class: `gaphas.tree.Tree`\n\nA Tree structure with the nodes stored in a depth-first order.\n\n```python\ntree = Tree()\ntree.add("node1")\ntree.add("node2", parent="node1")\n```\n\n#### Class: `gaphas.matrix.Matrix`\n\nAdds @observed messages on state changes to the cairo.Matrix.\n\n- xx (float): xx component of the affine transformation\n- yx (float): yx component of the affine transformation\n- xy (float): xy component of the affine transformation\n- yy (float): yy component of the affine transformation\n- x0 (float): X translation component of the affine transformation\n- y0 (float): Y translation component of the affine transformation\n\n```python\nmatrix = Matrix(1, 0, 0, 1, 0, 0)\n```\n\n#### Class: `gaphas.table.Table`\n\nTable is a storage class that can be used to store information, like one would in a database table, with indexes on the desired "columns." It includes indexing and is optimized for lookups.\n\n- columns (tuple): The columns of the table.\n- index (tuple):\n\n```python\nfrom collections import namedtuple\nC = namedtuple(\'C\', "foo bar baz")\ns = Table(C, (2,))\n```\n\n#### Class: `gaphas.quadtree.Quadtree`\n\nA quadtree is a tree data structure in which each internal node has up to four children. Quadtrees are most often used to partition a two\n\n- bounds (tuple): The boundaries of the quadtree (x, y, width, height).\n- capacity (int); The number of elements in one tree bucket; default is 10.\n\n```python\nqtree = Quadtree((0, 0, 100, 100))\n```\n\n#### Class: `gaphas.geometry.Rectangle`\n\nRectangle object which can be added (union), substituted (intersection), and points and rectangles can be tested to be in the rectangle.\n\n- x (int): X position of the rectangle.\n- y (int): Y position of the rectangle.\n- width (int): Width of the rectangle.\n- height (int): Hiehgt of the rectangle.\n\n```python\nrect = Rectangle(1, 1, 5, 5)\n```\n\n#### Decorator: @AsyncIO\n\nSchedule an idle handler at a given priority.\n\n- single (bool): Schedules the decorated function to be called only a single time.\n- timeout (int): The time between calls of the decorated function.\n- priority (int): The GLib.PRIORITY constant to set the event priority.\n\n```python\n\n@AsyncIO(single=True, timeout=60)\ndef c2(self):\n    print(\'idle-c2\', GLib.main_depth())\n```\n\n#### Decorator: @nonrecursive\n\nEnforces a function or method to not be executed recursively.\n   api/decorators\n\n```python\n\nclass A(object):\n    @nonrecursive\n    def a(self, x=1):\n        self.a(x+1)\n```\n\n#### Decorator: @recursive\n\nLimits the recursion for a specific function.\n\n- limit (int): The limit for the number of recursive loops a function can be called; default is 10000.\n\n```python\n@recursive(10)\ndef a(self, x=0):\n    self.a()\n```\n\n## :heart: Contributing\n\nThanks goes to these wonderful people ([emoji key](https://github.com/kentcdodds/all-contributors#emoji-key)):\n\n<!-- ALL-CONTRIBUTORS-LIST:START - Do not remove or modify this section -->\n<!-- prettier-ignore -->\n<table><tr><td align="center"><a href="https://github.com/amolenaar"><img src="https://avatars0.githubusercontent.com/u/96249?v=4" width="100px;" alt="Arjan Molenaar"/><br /><sub><b>Arjan Molenaar</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/commits?author=amolenaar" title="Code">💻</a> <a href="https://github.com/gaphor/gaphas/issues?q=author%3Aamolenaar" title="Bug reports">🐛</a> <a href="https://github.com/gaphor/gaphas/commits?author=amolenaar" title="Documentation">📖</a> <a href="#review-amolenaar" title="Reviewed Pull Requests">👀</a> <a href="#question-amolenaar" title="Answering Questions">💬</a> <a href="#plugin-amolenaar" title="Plugin/utility libraries">🔌</a></td><td align="center"><a href="https://ghuser.io/danyeaw"><img src="https://avatars1.githubusercontent.com/u/10014976?v=4" width="100px;" alt="Dan Yeaw"/><br /><sub><b>Dan Yeaw</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/commits?author=danyeaw" title="Code">💻</a> <a href="https://github.com/gaphor/gaphas/commits?author=danyeaw" title="Tests">⚠️</a> <a href="#review-danyeaw" title="Reviewed Pull Requests">👀</a> <a href="https://github.com/gaphor/gaphas/issues?q=author%3Adanyeaw" title="Bug reports">🐛</a> <a href="#question-danyeaw" title="Answering Questions">💬</a> <a href="#infra-danyeaw" title="Infrastructure (Hosting, Build-Tools, etc)">🚇</a> <a href="https://github.com/gaphor/gaphas/commits?author=danyeaw" title="Documentation">📖</a></td><td align="center"><a href="https://github.com/wrobell"><img src="https://avatars2.githubusercontent.com/u/105664?v=4" width="100px;" alt="wrobell"/><br /><sub><b>wrobell</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/commits?author=wrobell" title="Code">💻</a> <a href="https://github.com/gaphor/gaphas/commits?author=wrobell" title="Tests">⚠️</a> <a href="#review-wrobell" title="Reviewed Pull Requests">👀</a></td><td align="center"><a href="https://github.com/jlstevens"><img src="https://avatars3.githubusercontent.com/u/890576?v=4" width="100px;" alt="Jean-Luc Stevens"/><br /><sub><b>Jean-Luc Stevens</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/commits?author=jlstevens" title="Code">💻</a> <a href="https://github.com/gaphor/gaphas/issues?q=author%3Ajlstevens" title="Bug reports">🐛</a> <a href="https://github.com/gaphor/gaphas/commits?author=jlstevens" title="Documentation">📖</a></td><td align="center"><a href="http://www.franework.de"><img src="https://avatars1.githubusercontent.com/u/1144966?v=4" width="100px;" alt="Franz Steinmetz"/><br /><sub><b>Franz Steinmetz</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/commits?author=franzlst" title="Code">💻</a> <a href="https://github.com/gaphor/gaphas/issues?q=author%3Afranzlst" title="Bug reports">🐛</a></td><td align="center"><a href="https://github.com/adrianboguszewski"><img src="https://avatars3.githubusercontent.com/u/4547501?v=4" width="100px;" alt="Adrian Boguszewski"/><br /><sub><b>Adrian Boguszewski</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/commits?author=adrianboguszewski" title="Code">💻</a></td><td align="center"><a href="https://github.com/Rbelder"><img src="https://avatars3.githubusercontent.com/u/15119522?v=4" width="100px;" alt="Rico Belder"/><br /><sub><b>Rico Belder</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/issues?q=author%3ARbelder" title="Bug reports">🐛</a> <a href="#review-Rbelder" title="Reviewed Pull Requests">👀</a></td></tr><tr><td align="center"><a href="http://www.boduch.ca"><img src="https://avatars2.githubusercontent.com/u/114619?v=4" width="100px;" alt="Adam Boduch"/><br /><sub><b>Adam Boduch</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/issues?q=author%3Aadamboduch" title="Bug reports">🐛</a></td><td align="center"><a href="https://github.com/janettech"><img src="https://avatars3.githubusercontent.com/u/13398384?v=4" width="100px;" alt="Janet Jose"/><br /><sub><b>Janet Jose</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/commits?author=janettech" title="Documentation">📖</a></td></tr></table>\n\n<!-- ALL-CONTRIBUTORS-LIST:END -->\n\nThis project follows the [all-contributors](https://github.com/kentcdodds/all-contributors) specification. Contributions of any kind are welcome!\n\n1. Check for open issues or open a fresh issue to start a discussion around a feature idea or a bug.\n    There is a [first-timers-only](https://github.com/gaphor/gaphas/issues?utf8=%E2%9C%93&q=is%3Aissue+is%3Aopen+label%3Afirst-timers-only) tag for issues that should be ideal for people who are not very familiar with the codebase yet.\n2. Fork [the repository](https://github.com/gaphor/gaphas) on GitHub to    start making your changes to the **master**       branch (or branch off of it).\n3. Write a test which shows that the bug was fixed or that the feature\n   works as expected.\n4. Send a pull request and bug the maintainers until it gets merged and\n   published. :smile:\n\nSee [the contributing file](CONTRIBUTING.md)!\n\n## :copyright: License\n\nCopyright (C) Arjan Molenaar and Dan Yeaw\n\nLicensed under the [Apache License 2.0](LICENSE).\n\nSummary: You can do what you like with Gaphas, as long as you include the required notices. This permissive license contains a patent license from the contributors of the code.\n',
    'author': 'Arjan J. Molenaar',
    'author_email': 'gaphor@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gaphas.readthedocs.io/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



class Projection:
    """
    Projections are used to convert values from one space to another,
    e.g. from Canvas to Item space or visa versa.

    In order to be a Projection the ``value`` and ``strength``
    properties should be implemented and a method named ``variable()``
    should be present.

    Projections should inherit from this class.

    Projections may be nested.

    This default implementation projects a variable to it's own:

    >>> v = Variable(4.0)
    >>> v
    Variable(4, 20)
    >>> p = Projection(v)
    >>> p.value
    4.0
    >>> p.value = -1
    >>> p.value
    -1.0
    >>> v.value
    -1.0
    >>> p.strength
    20
    >>> p.variable()
    Variable(-1, 20)
    """

    def __init__(self, var):
        self._var = var

    def _set_value(self, value):
        self._var.value = value

    value = property(lambda s: s._var.value, _set_value)

    strength = property(lambda s: s._var.strength)

    def variable(self):
        """
        Return the variable owned by the projection.
        """
        return self._var

    def __float__(self):
        return float(self.variable()._value)

    def __str__(self):
        return f"{self.__class__.__name__}({self.variable()})"

    __repr__ = __str__


class VariableProjection(Projection):
    """
    Project a single `solver.Variable` to another space/coordinate system.

    The value has been set in the "other" coordinate system. A
    callback is executed when the value changes.

    It's a simple Variable-like class, following the Projection protocol:

    >>> def notify_me(val):
    ...     print('new value', val)
    >>> p = VariableProjection('var placeholder', 3.0, callback=notify_me)
    >>> p.value
    3.0
    >>> p.value = 6.5
    new value 6.5
    """

    def __init__(self, var, value, callback):
        self._var = var
        self._value = value
        self._callback = callback

    def _set_value(self, value):
        self._value = value
        self._callback(value)

    value = property(lambda s: s._value, _set_value)

    def variable(self):
        return self._var


class CanvasProjection:
    """
    Project a point as Canvas coordinates.  Although this is a
    projection, it behaves like a tuple with two Variables
    (Projections).

    >>> canvas = Canvas()
    >>> from gaphas.item import Element
    >>> a = Element()
    >>> canvas.add(a)
    >>> a.matrix.translate(30, 2)
    >>> canvas.request_matrix_update(a)
    >>> canvas.update_now()
    >>> canvas.get_matrix_i2c(a)
    cairo.Matrix(1, 0, 0, 1, 30, 2)
    >>> p = CanvasProjection(a.handles()[2].pos, a)
    >>> a.handles()[2].pos
    <Position object on (10, 10)>
    >>> p[0].value
    40.0
    >>> p[1].value
    12.0
    >>> p[0].value = 63
    >>> p._point
    <Position object on (33, 10)>

    When the variables are retrieved, new values are calculated.
    """

    def __init__(self, point, item):
        self._point = point
        self._item = item

    def _on_change_x(self, value):
        item = self._item
        self._px = value
        self._point.x.value, self._point.y.value = item.canvas.get_matrix_c2i(
            item
        ).transform_point(value, self._py)
        item.canvas.request_update(item, matrix=False)

    def _on_change_y(self, value):
        item = self._item
        self._py = value
        self._point.x.value, self._point.y.value = item.canvas.get_matrix_c2i(
            item
        ).transform_point(self._px, value)
        item.canvas.request_update(item, matrix=False)

    def _get_value(self):
        """
        Return two delegating variables. Each variable should contain
        a value attribute with the real value.
        """
        item = self._item
        x, y = self._point.x, self._point.y
        self._px, self._py = item.canvas.get_matrix_i2c(item).transform_point(x, y)
        return self._px, self._py

    pos = property(
        lambda self: list(
            map(
                VariableProjection,
                self._point,
                self._get_value(),
                (self._on_change_x, self._on_change_y),
            )
        )
    )

    def __getitem__(self, key):
        # Note: we can not use bound methods as callbacks, since that will
        #       cause pickle to fail.
        return self.pos[key]

    def __iter__(self):
        return iter(self.pos)



"""Defines aspects for Items.

Aspects form intermediate items between tools and items.
"""
import warnings
from functools import singledispatch as real_singledispatch

from gi.repository import Gdk

from gaphas.item import Element, Item


def singledispatch(func):
    """
    Wrapper around singledispatch(), with an extra compatibility function
    so code will not break when upgrading from 1.0 to 1.1.
    """
    wrapped = real_singledispatch(func)

    def when_type(*types):
        if not types:
            raise TypeError("should provide at least one type")
        warnings.warn(
            "when_type: is deprecated, use `register` instead",
            category=DeprecationWarning,
            stacklevel=2,
        )

        def wrapper_for_types(func):
            for cls in types:
                wrapped.register(cls, func)
            return func

        return wrapper_for_types

    wrapped.when_type = when_type
    return wrapped


class ItemFinder:
    """
    Find an item on the canvas.
    """

    def __init__(self, view):
        self.view = view

    def get_item_at_point(self, pos):
        item, handle = self.view.get_handle_at_point(pos)
        return item or self.view.get_item_at_point(pos)


Finder = singledispatch(ItemFinder)


class ItemSelection:
    """
    A role for items. When dealing with selection.

    Behaviour can be overridden by applying the @aspect decorator
    to a subclass.
    """

    def __init__(self, item, view):
        self.item = item
        self.view = view

    def select(self):
        """
        Set selection on the view.
        """
        self.view.focused_item = self.item

    def unselect(self):
        self.view.focused_item = None
        self.view.unselect_item(self.item)


Selection = singledispatch(ItemSelection)


class ItemInMotion:
    """
    Aspect for dealing with motion on an item.

    In this case the item is moved.
    """

    def __init__(self, item, view):
        self.item = item
        self.view = view
        self.last_x, self.last_y = None, None

    def start_move(self, pos):
        self.last_x, self.last_y = pos

    def move(self, pos):
        """
        Move the item. x and y are in view coordinates.
        """
        item = self.item
        view = self.view
        v2i = view.get_matrix_v2i(item)

        x, y = pos
        dx, dy = x - self.last_x, y - self.last_y
        dx, dy = v2i.transform_distance(dx, dy)
        self.last_x, self.last_y = x, y

        item.matrix.translate(dx, dy)
        item.canvas.request_matrix_update(item)

    def stop_move(self):
        pass


InMotion = singledispatch(ItemInMotion)


class ItemHandleFinder:
    """
    Deals with the task of finding handles.
    """

    def __init__(self, item, view):
        self.item = item
        self.view = view

    def get_handle_at_point(self, pos):
        return self.view.get_handle_at_point(pos)


HandleFinder = singledispatch(ItemHandleFinder)


class ItemHandleSelection:
    """
    Deal with selection of the handle.
    """

    def __init__(self, item, handle, view):
        self.item = item
        self.handle = handle
        self.view = view

    def select(self):
        pass

    def unselect(self):
        pass


HandleSelection = singledispatch(ItemHandleSelection)


@HandleSelection.register(Element)
class ElementHandleSelection(ItemHandleSelection):
    CURSORS = ("nw-resize", "ne-resize", "se-resize", "sw-resize")

    def select(self):
        index = self.item.handles().index(self.handle)
        if index < 4:
            display = self.view.get_display()
            cursor = Gdk.Cursor.new_from_name(display, self.CURSORS[index])
            self.view.get_window().set_cursor(cursor)

    def unselect(self):
        from .view import DEFAULT_CURSOR

        cursor = Gdk.Cursor(DEFAULT_CURSOR)
        self.view.get_window().set_cursor(cursor)


class ItemHandleInMotion:
    """
    Move a handle (role is applied to the handle)
    """

    GLUE_DISTANCE = 10

    def __init__(self, item, handle, view):
        self.item = item
        self.handle = handle
        self.view = view
        self.last_x, self.last_y = None, None

    def start_move(self, pos):
        self.last_x, self.last_y = pos
        canvas = self.item.canvas

        cinfo = canvas.get_connection(self.handle)
        if cinfo:
            canvas.solver.remove_constraint(cinfo.constraint)

    def move(self, pos):
        item = self.item
        handle = self.handle
        view = self.view

        v2i = view.get_matrix_v2i(item)

        x, y = v2i.transform_point(*pos)

        self.handle.pos = (x, y)

        sink = self.glue(pos)

        # do not request matrix update as matrix recalculation will be
        # performed due to item normalization if required
        item.request_update(matrix=False)

        return sink

    def stop_move(self):
        pass

    def glue(self, pos, distance=GLUE_DISTANCE):
        """
        Glue to an item near a specific point.

        Returns a ConnectionSink or None.
        """
        item = self.item
        handle = self.handle
        view = self.view

        if not handle.connectable:
            return None

        connectable, port, glue_pos = view.get_port_at_point(
            pos, distance=distance, exclude=(item,)
        )

        # check if item and found item can be connected on closest port
        if port is not None:
            assert connectable is not None

            connector = Connector(self.item, self.handle)
            sink = ConnectionSink(connectable, port)

            if connector.allow(sink):
                # transform coordinates from view space to the item
                # space and update position of item's handle
                v2i = view.get_matrix_v2i(item).transform_point
                handle.pos = v2i(*glue_pos)
                return sink
        return None


HandleInMotion = singledispatch(ItemHandleInMotion)


class ItemConnector:
    """Connect or disconnect an item's handle to another item or port.

    """

    GLUE_DISTANCE = 10  # Glue distance in view points

    def __init__(self, item, handle):
        self.item = item
        self.handle = handle

    def allow(self, sink):
        return True

    def glue(self, sink):
        """
        Glue the Connector handle on the sink's port.
        """
        handle = self.handle
        item = self.item
        matrix = item.canvas.get_matrix_i2i(item, sink.item)
        pos = matrix.transform_point(*handle.pos)
        gluepos, dist = sink.port.glue(pos)
        matrix.invert()
        handle.pos = matrix.transform_point(*gluepos)

    def connect(self, sink):
        """
        Connect the handle to a sink (item, port).

        Note that connect() also takes care of disconnecting in case a
        handle is reattached to another element.
        """

        cinfo = self.item.canvas.get_connection(self.handle)

        # Already connected? disconnect first.
        if cinfo:
            self.disconnect()

        if not self.allow(sink):
            return

        self.glue(sink)

        self.connect_handle(sink)

    def connect_handle(self, sink, callback=None):
        """
        Create constraint between handle of a line and port of
        connectable item.

        :Parameters:
         sink
            Connectable item and port.
         callback
            Function to be called on disconnection.
        """
        canvas = self.item.canvas
        handle = self.handle
        item = self.item

        constraint = sink.port.constraint(canvas, item, handle, sink.item)

        canvas.connect_item(
            item, handle, sink.item, sink.port, constraint, callback=callback
        )

    def disconnect(self):
        """
        Disconnect the handle from the attached element.
        """
        self.item.canvas.disconnect_item(self.item, self.handle)


Connector = singledispatch(ItemConnector)


class ItemConnectionSink:
    """Makes an item a sink.

    A sink is another item that an item's handle is connected to like a
    connectable item or port.

    """

    def __init__(self, item, port):
        self.item = item
        self.port = port

    def find_port(self, pos):
        """
        Glue to the closest item on the canvas.
        If the item can connect, it returns a port.
        """
        port = None
        max_dist = 10e6
        for p in self.item.ports():
            pg, d = p.glue(pos)
            if d >= max_dist:
                continue
            port = p
            max_dist = d

        return port


ConnectionSink = singledispatch(ItemConnectionSink)


#
# Painter aspects
#


class ItemPaintFocused:
    """
    Paints on top of all items, just for the focused item and only
    when it's hovered (see gaphas.painter.FocusedItemPainter)
    """

    def __init__(self, item, view):
        self.item = item
        self.view = view

    def paint(self, context):
        pass


PaintFocused = singledispatch(ItemPaintFocused)



"""
Tools provide interactive behavior to a `View` by handling specific
events sent by view.

Some of implemented tools are

`HoverTool`
    make the item under the mouse cursor the "hovered item"

`ItemTool`
    handle selection and movement of items

`HandleTool`
    handle selection and movement of handles

`RubberbandTool`
    for rubber band selection of multiple items

`PanTool`
    for easily moving the canvas around

`PlacementTool`
    for placing items on the canvas

The tools are chained with `ToolChain` class (it is a tool as well),
which allows to combine functionality provided by different tools.

Tools can handle events in different ways

- event can be ignored
- tool can handle the event (obviously)
"""
from gi.repository import Gdk, Gtk

from gaphas.aspect import (
    Connector,
    Finder,
    HandleFinder,
    HandleInMotion,
    HandleSelection,
    InMotion,
    Selection,
)
from gaphas.canvas import Context

DEBUG_TOOL_CHAIN = False

Event = Context


class Tool:
    """
    Base class for a tool. This class
    A word on click events:

    Mouse (pointer) button click. A button press is normally followed
    by a button release. Double and triple clicks should work together
    with the button methods.

    A single click is emitted as:
            on_button_press
            on_button_release

    In case of a double click:
            on_button_press (x 2)
            on_double_click
            on_button_release

    In case of a triple click:
            on_button_press (x 3)
            on_triple_click
            on_button_release
    """

    # Custom events:
    GRAB = -100
    UNGRAB = -101

    # Map GDK events to tool methods
    EVENT_HANDLERS = {
        Gdk.EventType.BUTTON_PRESS: "on_button_press",
        Gdk.EventType.BUTTON_RELEASE: "on_button_release",
        Gdk.EventType._2BUTTON_PRESS: "on_double_click",
        Gdk.EventType._3BUTTON_PRESS: "on_triple_click",
        Gdk.EventType.MOTION_NOTIFY: "on_motion_notify",
        Gdk.EventType.KEY_PRESS: "on_key_press",
        Gdk.EventType.KEY_RELEASE: "on_key_release",
        Gdk.EventType.SCROLL: "on_scroll",
        # Custom events:
        GRAB: "on_grab",
        UNGRAB: "on_ungrab",
    }

    # Those events force the tool to release the grabbed tool.
    FORCE_UNGRAB_EVENTS = (Gdk.EventType._2BUTTON_PRESS, Gdk.EventType._3BUTTON_PRESS)

    def __init__(self, view=None):
        self.view = view

    def set_view(self, view):
        self.view = view

    def _dispatch(self, event):
        """
        Deal with the event. The event is dispatched to a specific
        handler for the event type.
        """
        handler = self.EVENT_HANDLERS.get(event.type)
        if handler:
            try:
                h = getattr(self, handler)
            except AttributeError:
                pass  # No handler
            else:
                return bool(h(event))
        return False

    def handle(self, event):
        return self._dispatch(event)

    def draw(self, context):
        """
        Some tools (such as Rubberband selection) may need to draw
        something on the canvas. This can be done through the draw()
        method. This is called after all items are drawn.
        The context contains the following fields:

        - context: the render context (contains context.view and context.cairo)
        - cairo: the Cairo drawing context
        """
        pass


class ToolChain(Tool):
    """
    A ToolChain can be used to chain tools together, for example
    HoverTool, HandleTool, SelectionTool.

    The grabbed item is bypassed in case a double or triple click
    event is received. Should make sure this doesn't end up in
    dangling states.
    """

    def __init__(self, view=None):
        super().__init__(view)
        self._tools = []
        self._grabbed_tool = None

    def set_view(self, view):
        self.view = view
        for tool in self._tools:
            tool.set_view(self.view)

    def append(self, tool):
        """
        Append a tool to the chain. Self is returned.
        """
        self._tools.append(tool)
        tool.view = self.view
        return self

    def grab(self, tool):
        if not self._grabbed_tool:
            if DEBUG_TOOL_CHAIN:
                print("Grab tool", tool)
            # Send grab event
            event = Event(type=Tool.GRAB)
            tool.handle(event)
            self._grabbed_tool = tool

    def ungrab(self, tool):
        if self._grabbed_tool is tool:
            if DEBUG_TOOL_CHAIN:
                print("UNgrab tool", self._grabbed_tool)
            # Send ungrab event
            event = Event(type=Tool.UNGRAB)
            tool.handle(event)
            self._grabbed_tool = None

    def validate_grabbed_tool(self, event):
        """
        Check if it's valid to have a grabbed tool on an event. If not
        the grabbed tool will be released.
        """
        if event.type in self.FORCE_UNGRAB_EVENTS:
            self.ungrab(self._grabbed_tool)

    def handle(self, event):
        """
        Handle the event by calling each tool until the event is
        handled or grabbed.

        If a tool is returning True on a button press event, the
        motion and button release events are also passed to this
        """
        handler = self.EVENT_HANDLERS.get(event.type)

        self.validate_grabbed_tool(event)

        if self._grabbed_tool and handler:
            try:
                return self._grabbed_tool.handle(event)
            finally:
                if event.type == Gdk.EventType.BUTTON_RELEASE:
                    self.ungrab(self._grabbed_tool)
        else:
            for tool in self._tools:
                if DEBUG_TOOL_CHAIN:
                    print("tool", tool)
                rt = tool.handle(event)
                if rt:
                    if event.type == Gdk.EventType.BUTTON_PRESS:
                        self.view.grab_focus()
                        self.grab(tool)
                    return rt

    def draw(self, context):
        if self._grabbed_tool:
            self._grabbed_tool.draw(context)


class HoverTool(Tool):
    """
    Make the item under the mouse cursor the "hovered item".
    """

    def on_motion_notify(self, event):
        view = self.view
        pos = event.get_coords()[1:]
        view.hovered_item = Finder(view).get_item_at_point(pos)


class ItemTool(Tool):
    """
    ItemTool does selection and dragging of items. On a button click,
    the currently "hovered item" is selected. If CTRL or SHIFT are
    pressed, already selected items remain selected. The last selected
    item gets the focus (e.g. receives key press events).

    The roles used are Selection (select, unselect) and InMotion (move).
    """

    def __init__(self, view=None, buttons=(1,)):
        super().__init__(view)
        self._buttons = buttons
        self._movable_items = set()

    def get_item(self):
        return self.view.hovered_item

    def movable_items(self):
        """
        Filter the items that should eventually be moved.

        Returns InMotion aspects for the items.
        """
        view = self.view
        get_ancestors = view.canvas.get_ancestors
        selected_items = set(view.selected_items)
        for item in selected_items:
            # Do not move subitems of selected items
            if not set(get_ancestors(item)).intersection(selected_items):
                yield InMotion(item, view)

    def on_button_press(self, event):
        ### TODO: make keys configurable
        view = self.view
        item = self.get_item()

        if event.get_button()[1] not in self._buttons:
            return False

        # Deselect all items unless CTRL or SHIFT is pressed
        # or the item is already selected.
        if not (
            event.get_state()[1]
            & (Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK)
            or item in view.selected_items
        ):
            del view.selected_items

        if item:
            if (
                view.hovered_item in view.selected_items
                and event.get_state()[1] & Gdk.ModifierType.CONTROL_MASK
            ):
                selection = Selection(item, view)
                selection.unselect()
            else:
                selection = Selection(item, view)
                selection.select()
                self._movable_items.clear()
            return True

    def on_button_release(self, event):
        if event.get_button()[1] not in self._buttons:
            return False
        for inmotion in self._movable_items:
            inmotion.stop_move()
        self._movable_items.clear()
        return True

    def on_motion_notify(self, event):
        """
        Normally do nothing.
        If a button is pressed move the items around.
        """
        if event.get_state()[1] & Gdk.EventMask.BUTTON_PRESS_MASK:

            if not self._movable_items:
                self._movable_items = set(self.movable_items())
                for inmotion in self._movable_items:
                    inmotion.start_move(event.get_coords()[1:])

            for inmotion in self._movable_items:
                inmotion.move(event.get_coords()[1:])

            return True


class HandleTool(Tool):
    """
    Tool for moving handles around.

    By default this tool does not provide connecting handles to
    another item (see `ConnectHandleTool`).
    """

    def __init__(self, view=None):
        super().__init__(view)
        self.grabbed_handle = None
        self.grabbed_item = None
        self.motion_handle = None

    def grab_handle(self, item, handle):
        """
        Grab a specific handle. This can be used from the
        PlacementTool to set the state of the handle tool.
        """
        assert item is None and handle is None or handle in item.handles()
        self.grabbed_item = item
        self.grabbed_handle = handle

        selection = HandleSelection(item, handle, self.view)
        selection.select()

    def ungrab_handle(self):
        """
        Reset grabbed_handle and grabbed_item.
        """
        item = self.grabbed_item
        handle = self.grabbed_handle
        self.grabbed_handle = None
        self.grabbed_item = None
        if handle:
            selection = HandleSelection(item, handle, self.view)
            selection.unselect()

    def on_button_press(self, event):
        """
        Handle button press events. If the (mouse) button is pressed
        on top of a Handle (item.Handle), that handle is grabbed and
        can be dragged around.
        """
        view = self.view

        item, handle = HandleFinder(view.hovered_item, view).get_handle_at_point(
            event.get_coords()[1:]
        )

        if handle:
            # Deselect all items unless CTRL or SHIFT is pressed
            # or the item is already selected.
            ### TODO: duplicate from ItemTool
            if not (
                event.get_state()[1]
                & (Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK)
                or view.hovered_item in view.selected_items
            ):
                del view.selected_items
            ###/
            view.hovered_item = item
            view.focused_item = item

            self.motion_handle = None

            self.grab_handle(item, handle)

            return True

    def on_button_release(self, event):
        """
        Release a grabbed handle.
        """
        # queue extra redraw to make sure the item is drawn properly
        grabbed_handle, grabbed_item = self.grabbed_handle, self.grabbed_item

        if self.motion_handle:
            self.motion_handle.stop_move()
            self.motion_handle = None

        self.ungrab_handle()

        if grabbed_handle:
            grabbed_item.request_update()
        return True

    def on_motion_notify(self, event):
        """
        Handle motion events. If a handle is grabbed: drag it around,
        else, if the pointer is over a handle, make the owning item
        the hovered-item.
        """
        view = self.view
        if (
            self.grabbed_handle
            and event.get_state()[1] & Gdk.EventMask.BUTTON_PRESS_MASK
        ):
            canvas = view.canvas
            item = self.grabbed_item
            handle = self.grabbed_handle
            pos = event.get_coords()[1:]

            if not self.motion_handle:
                self.motion_handle = HandleInMotion(item, handle, self.view)
                self.motion_handle.start_move(pos)
            self.motion_handle.move(pos)

            return True


class RubberbandTool(Tool):
    def __init__(self, view=None):
        super().__init__(view)
        self.x0, self.y0, self.x1, self.y1 = 0, 0, 0, 0

    def on_button_press(self, event):
        self.x0, self.y0 = event.get_coords()[1:]
        self.x1, self.y1 = event.get_coords()[1:]
        return True

    def on_button_release(self, event):
        self.queue_draw(self.view)
        x0, y0, x1, y1 = self.x0, self.y0, self.x1, self.y1
        self.view.select_in_rectangle(
            (min(x0, x1), min(y0, y1), abs(x1 - x0), abs(y1 - y0))
        )
        return True

    def on_motion_notify(self, event):
        if event.get_state()[1] & Gdk.EventMask.BUTTON_PRESS_MASK:
            view = self.view
            self.queue_draw(view)
            self.x1, self.y1 = event.get_coords()[1:]
            self.queue_draw(view)
            return True

    def queue_draw(self, view):
        x0, y0, x1, y1 = self.x0, self.y0, self.x1, self.y1
        view.queue_draw_area(min(x0, x1), min(y0, y1), abs(x1 - x0), abs(y1 - y0))

    def draw(self, context):
        cr = context.cairo
        x0, y0, x1, y1 = self.x0, self.y0, self.x1, self.y1
        cr.rectangle(min(x0, x1), min(y0, y1), abs(x1 - x0), abs(y1 - y0))
        cr.set_source_rgba(0.9, 0.9, 0.9, 0.3)
        cr.fill_preserve()
        cr.set_line_width(2.0)
        cr.set_dash((7.0, 5.0), 0)
        cr.set_source_rgba(0.5, 0.5, 0.7, 0.7)
        cr.stroke()


PAN_MASK = (
    Gdk.ModifierType.SHIFT_MASK
    | Gdk.ModifierType.MOD1_MASK
    | Gdk.ModifierType.CONTROL_MASK
)
PAN_VALUE = 0


class PanTool(Tool):
    """
    Captures drag events with the middle mouse button and uses them to
    translate the canvas within the view. Trumps the ZoomTool, so
    should be placed later in the ToolChain.
    """

    def __init__(self, view=None):
        super().__init__(view)
        self.x0, self.y0 = 0, 0
        self.speed = 10

    def on_button_press(self, event):
        if event.get_state()[1] & PAN_MASK != PAN_VALUE:
            return False
        if event.get_button()[1] == 2:
            self.x0, self.y0 = event.get_coords()[1:]
            return True

    def on_button_release(self, event):
        self.x0, self.y0 = event.get_coords()[1:]
        return True

    def on_motion_notify(self, event):
        if event.get_state()[1] & Gdk.ModifierType.BUTTON2_MASK:
            view = self.view
            self.x1, self.y1 = event.get_coords()[1:]
            dx = self.x1 - self.x0
            dy = self.y1 - self.y0
            view._matrix.translate(dx / view._matrix[0], dy / view._matrix[3])
            # Make sure everything's updated
            view.request_update((), view._canvas.get_all_items())
            self.x0 = self.x1
            self.y0 = self.y1
            return True

    def on_scroll(self, event):
        # Ensure no modifiers
        if event.get_state()[1] & PAN_MASK != PAN_VALUE:
            return False
        view = self.view
        direction = event.get_scroll_direction()[1]
        if direction == Gdk.ScrollDirection.LEFT:
            view._matrix.translate(self.speed / view._matrix[0], 0)
        elif direction == Gdk.ScrollDirection.RIGHT:
            view._matrix.translate(-self.speed / view._matrix[0], 0)
        elif direction == Gdk.ScrollDirection.UP:
            view._matrix.translate(0, self.speed / view._matrix[3])
        elif direction == Gdk.ScrollDirection.DOWN:
            view._matrix.translate(0, -self.speed / view._matrix[3])
        view.request_update((), view._canvas.get_all_items())
        return True


ZOOM_MASK = (
    Gdk.ModifierType.CONTROL_MASK
    | Gdk.ModifierType.SHIFT_MASK
    | Gdk.ModifierType.MOD1_MASK
)
ZOOM_VALUE = Gdk.ModifierType.CONTROL_MASK


class ZoomTool(Tool):
    """Tool for zooming.

    Uses two different user inputs to zoom:
    - Ctrl + middle-mouse dragging in the up-down direction.
    - Ctrl + mouse-wheel

    """

    def __init__(self, view=None):
        super().__init__(view)
        self.x0, self.y0 = 0, 0
        self.lastdiff = 0

    def on_button_press(self, event):
        if (
            event.get_button()[1] == 2
            and event.get_state()[1] & ZOOM_MASK == ZOOM_VALUE
        ):
            pos = event.get_coords()[1:]
            self.x0 = pos[0]
            self.y0 = pos[1]
            self.lastdiff = 0
            return True

    def on_button_release(self, event):
        self.lastdiff = 0
        return True

    def on_motion_notify(self, event):
        if (
            event.get_state()[1] & ZOOM_MASK != ZOOM_VALUE
            or not event.get_state()[1] & Gdk.ModifierType.BUTTON2_MASK
        ):
            return

        view = self.view
        pos = event.get_coords()[1:]
        dy = pos[1] - self.y0

        sx = view._matrix[0]
        sy = view._matrix[3]
        ox = (view._matrix[4] - self.x0) / sx
        oy = (view._matrix[5] - self.y0) / sy

        if abs(dy - self.lastdiff) > 20:
            factor = 1.0 / 0.9 if dy - self.lastdiff < 0 else 0.9
            m = view.matrix
            m.translate(-ox, -oy)
            m.scale(factor, factor)
            m.translate(+ox, +oy)

            # Make sure everything's updated
            view.request_update((), view._canvas.get_all_items())

            self.lastdiff = dy
        return True

    def on_scroll(self, event):
        if event.get_state()[1] & Gdk.ModifierType.CONTROL_MASK:
            view = self.view
            sx = view._matrix[0]
            sy = view._matrix[3]
            pos = event.get_coords()[1:]
            ox = (view._matrix[4] - pos[0]) / sx
            oy = (view._matrix[5] - pos[1]) / sy
            factor = 0.9
            if event.get_scroll_direction()[1] == Gdk.ScrollDirection.UP:
                factor = 1.0 / factor
            view._matrix.translate(-ox, -oy)
            view._matrix.scale(factor, factor)
            view._matrix.translate(+ox, +oy)
            # Make sure everything's updated
            view.request_update((), view._canvas.get_all_items())
            return True


class PlacementTool(Tool):
    def __init__(self, view, factory, handle_tool, handle_index):
        super().__init__(view)
        self._factory = factory
        self.handle_tool = handle_tool
        handle_tool.set_view(view)
        self._handle_index = handle_index
        self._new_item = None
        self.grabbed_handle = None

    # handle_tool = property(lambda s: s._handle_tool, doc="Handle tool")
    handle_index = property(
        lambda s: s._handle_index, doc="Index of handle to be used by handle_tool"
    )
    new_item = property(lambda s: s._new_item, doc="The newly created item")

    def on_button_press(self, event):
        view = self.view
        canvas = view.canvas
        pos = event.get_coords()[1:]
        new_item = self._create_item(pos)
        # Enforce matrix update, as a good matrix is required for the handle
        # positioning:
        canvas.get_matrix_i2c(new_item, calculate=True)

        self._new_item = new_item
        view.focused_item = new_item

        h = new_item.handles()[self._handle_index]
        if h.movable:
            self.handle_tool.grab_handle(new_item, h)
            self.grabbed_handle = h
        return True

    def _create_item(self, pos, **kw):
        view = self.view
        item = self._factory(**kw)
        x, y = view.get_matrix_v2i(item).transform_point(*pos)
        item.matrix.translate(x, y)
        return item

    def on_button_release(self, event):
        if self.grabbed_handle:
            self.handle_tool.on_button_release(event)
            self.grabbed_handle = None
        self._new_item = None
        return True

    def on_motion_notify(self, event):
        if self.grabbed_handle:
            return self.handle_tool.on_motion_notify(event)
        else:
            # act as if the event is handled if we have a new item
            return bool(self._new_item)


class TextEditTool(Tool):
    """
    Demo of a text edit tool (just displays a text edit box at the
    cursor position.
    """

    def __init__(self, view=None):
        super().__init__(view)

    def on_double_click(self, event):
        """
        Create a popup window with some editable text.
        """
        window = Gtk.Window()
        window.set_property("decorated", False)
        window.set_resize_mode(Gtk.ResizeMode.IMMEDIATE)
        # window.set_modal(True)
        window.set_parent_window(self.view.get_window())
        buffer = Gtk.TextBuffer()
        text_view = Gtk.TextView()
        text_view.set_buffer(buffer)
        text_view.show()
        window.add(text_view)
        rect = Gdk.Rectangle()
        pos = event.get_coords()[1:]
        rect.x, rect.y, rect.width, rect.height = int(pos[0]), int(pos[1]), 50, 50
        window.size_allocate(rect)
        # window.move(int(event.x), int(event.y))
        cursor_pos = self.view.get_toplevel().get_screen().get_display().get_pointer()
        window.move(cursor_pos[1], cursor_pos[2])
        window.connect("focus-out-event", self._on_focus_out_event, buffer)
        text_view.connect("key-press-event", self._on_key_press_event, buffer)
        # text_view.set_size_request(50, 50)
        window.show()
        # text_view.grab_focus()
        # window.set_uposition(event.x, event.y)
        # window.focus
        return True

    def _on_key_press_event(self, widget, event, buffer):
        # if event.keyval == Gdk.KEY_Return:
        #     widget.get_toplevel().destroy()
        if event.get_keyval()[1] == Gdk.KEY_Escape:
            widget.get_toplevel().destroy()

    def _on_focus_out_event(self, widget, event, buffer):
        widget.destroy()


class ConnectHandleTool(HandleTool):
    """
    Tool for connecting two items.

    There are two items involved. Handle of connecting item (usually a
    line) is being dragged by an user towards another item (item in
    short). Port of an item is found by the tool and connection is
    established by creating a constraint between line's handle and
    item's port.
    """

    def glue(self, item, handle, vpos):
        """
        Perform a small glue action to ensure the handle is at a
        proper location for connecting.
        """
        if self.motion_handle:
            return self.motion_handle.glue(vpos)
        else:
            return HandleInMotion(item, handle, self.view).glue(vpos)

    def connect(self, item, handle, vpos):
        """
        Connect a handle of a item to connectable item.

        Connectable item is found by `ConnectHandleTool.glue` method.

        :Parameters:
         item
            Connecting item.
         handle
            Handle of connecting item.
         vpos
            Position to connect to (or near at least)
        """
        connector = Connector(item, handle)

        # find connectable item and its port
        sink = self.glue(item, handle, vpos)

        # no new connectable item, then diconnect and exit
        if sink:
            connector.connect(sink)
        else:
            cinfo = item.canvas.get_connection(handle)
            if cinfo:
                connector.disconnect()

    def on_button_release(self, event):
        view = self.view
        item = self.grabbed_item
        handle = self.grabbed_handle
        try:
            if handle and handle.connectable:
                pos = event.get_coords()[1:]
                self.connect(item, handle, pos)
        finally:
            return super().on_button_release(event)


def DefaultTool(view=None):
    """
    The default tool chain build from HoverTool, ItemTool and HandleTool.
    """
    return (
        ToolChain(view)
        .append(HoverTool())
        .append(ConnectHandleTool())
        .append(PanTool())
        .append(ZoomTool())
        .append(ItemTool())
        .append(TextEditTool())
        .append(RubberbandTool())
    )



"""
Basic connectors such as Ports and Handles.
"""

import functools
import warnings

from gaphas.constraint import LineConstraint, PositionConstraint
from gaphas.geometry import distance_line_point, distance_point_point
from gaphas.solver import NORMAL, solvable
from gaphas.state import observed, reversible_property


def deprecated(message, since):
    def _deprecated(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            warnings.warn(
                f"{func.__name__}: {message}", category=DeprecationWarning, stacklevel=2
            )
            return func(*args, **kwargs)

        return wrapper

    return _deprecated


class Position:
    """
    A point constructed of two `Variable`'s.

    >>> vp = Position((3, 5))
    >>> vp.x, vp.y
    (Variable(3, 20), Variable(5, 20))
    >>> vp.pos
    (Variable(3, 20), Variable(5, 20))
    >>> vp[0], vp[1]
    (Variable(3, 20), Variable(5, 20))
    """

    x = solvable(varname="_v_x")
    y = solvable(varname="_v_y")

    def __init__(self, pos, strength=NORMAL):
        self.x, self.y = pos
        self.x.strength = strength
        self.y.strength = strength

    def _set_pos(self, pos):
        """
        Set handle position (Item coordinates).
        """
        self.x, self.y = pos

    pos = property(lambda s: (s.x, s.y), _set_pos)

    def __str__(self):
        return f"<{self.__class__.__name__} object on ({self.x}, {self.y})>"

    __repr__ = __str__

    def __getitem__(self, index):
        """
        Shorthand for returning the x(0) or y(1) component of the point.

        >>> h = Position((3, 5))
        >>> h[0]
        Variable(3, 20)
        >>> h[1]
        Variable(5, 20)
        """
        return (self.x, self.y)[index]


class Handle:
    """
    Handles are used to support modifications of Items.

    If the handle is connected to an item, the ``connected_to``
    property should refer to the item. A ``disconnect`` handler should
    be provided that handles all disconnect behaviour (e.g. clean up
    constraints and ``connected_to``).

    Note for those of you that use the Pickle module to persist a
    canvas: The property ``disconnect`` should contain a callable
    object (with __call__() method), so the pickle handler can also
    pickle that. Pickle is not capable of pickling ``instancemethod``
    or ``function`` objects.
    """

    def __init__(self, pos=(0, 0), strength=NORMAL, connectable=False, movable=True):
        self._pos = Position(pos, strength)
        self._connectable = connectable
        self._movable = movable
        self._visible = True

    def _set_pos(self, pos):
        """
        Shortcut for ``handle.pos.pos = pos``

        >>> h = Handle((10, 10))
        >>> h.pos
        <Position object on (10, 10)>
        >>> h.pos = (20, 15)
        >>> h.pos
        <Position object on (20, 15)>
        """
        self._pos.pos = pos

    pos = property(lambda s: s._pos, _set_pos)

    @deprecated("Use Handle.pos", "1.1.0")
    def _set_x(self, x):
        """
        Shortcut for ``handle.pos.x = x``
        """
        self._pos.x = x

    @deprecated("Use Handle.pos.x", "1.1.0")
    def _get_x(self):
        return self._pos.x

    x = property(_get_x, _set_x)

    @deprecated("Use Handle.pos", "1.1.0")
    def _set_y(self, y):
        """
        Shortcut for ``handle.pos.y = y``
        """
        self._pos.y = y

    @deprecated("Use Handle.pos.y", "1.1.0")
    def _get_y(self):
        return self._pos.y

    y = property(_get_y, _set_y)

    @observed
    def _set_connectable(self, connectable):
        self._connectable = connectable

    connectable = reversible_property(lambda s: s._connectable, _set_connectable)

    @observed
    def _set_movable(self, movable):
        self._movable = movable

    movable = reversible_property(lambda s: s._movable, _set_movable)

    @observed
    def _set_visible(self, visible):
        self._visible = visible

    visible = reversible_property(lambda s: s._visible, _set_visible)

    def __str__(self):
        return f"<{self.__class__.__name__} object on ({self._pos.x}, {self._pos.y})>"

    __repr__ = __str__


class Port:
    """Port connectable part of an item.

    The Item's handle connects to a port.

    """

    def __init__(self):
        super().__init__()

        self._connectable = True

    @observed
    def _set_connectable(self, connectable):
        self._connectable = connectable

    connectable = reversible_property(lambda s: s._connectable, _set_connectable)

    def glue(self, pos):
        """
        Get glue point on the port and distance to the port.
        """
        raise NotImplemented("Glue method not implemented")

    def constraint(self, canvas, item, handle, glue_item):
        """
        Create connection constraint between item's handle and glue item.
        """
        raise NotImplemented("Constraint method not implemented")


class LinePort(Port):
    """
    Port defined as a line between two handles.
    """

    def __init__(self, start, end):
        super().__init__()

        self.start = start
        self.end = end

    def glue(self, pos):
        """
        Get glue point on the port and distance to the port.

        >>> p1, p2 = (0.0, 0.0), (100.0, 100.0)
        >>> port = LinePort(p1, p2)
        >>> port.glue((50, 50))
        ((50.0, 50.0), 0.0)
        >>> port.glue((0, 10))
        ((5.0, 5.0), 7.0710678118654755)
        """
        d, pl = distance_line_point(self.start, self.end, pos)
        return pl, d

    def constraint(self, canvas, item, handle, glue_item):
        """
        Create connection line constraint between item's handle and
        the port.
        """
        line = canvas.project(glue_item, self.start, self.end)
        point = canvas.project(item, handle.pos)
        return LineConstraint(line, point)


class PointPort(Port):
    """
    Port defined as a point.
    """

    def __init__(self, point):
        super().__init__()
        self.point = point

    def glue(self, pos):
        """
        Get glue point on the port and distance to the port.

        >>> h = Handle((10, 10))
        >>> port = PointPort(h.pos)
        >>> port.glue((10, 0))
        (<Position object on (10, 10)>, 10.0)
        """
        d = distance_point_point(self.point, pos)
        return self.point, d

    def constraint(self, canvas, item, handle, glue_item):
        """
        Return connection position constraint between item's handle
        and the port.
        """
        origin = canvas.project(glue_item, self.point)
        point = canvas.project(item, handle.pos)
        c = PositionConstraint(origin, point)
        return c  # PositionConstraint(origin, point)
